import fruit.modules.step as step
import fruit.modules.job as job
import yaml
import click
import fruit.config as config


class fruitError(Exception):
    pass


class JobNotFoundError(Exception):
    """The requested job cannot be found"""
    pass

class TargetParseError(Exception):
    """Error while parsing a target"""
    pass

class Target(object):
    """
    Target class for build target functionalities.
    """

    name: str = ""  # Name of the target (build-nane)
    desc: str = ""  # Description of the target
    glob: dict = None  # List of global variables in the target
    jobs: list = None  # List of pipeline steps
    lists: list = None  # List comprehensions for commands

    def __init__(self, yaml_string: dict):
        """
        Create a new target from a parsed yaml dictionary.
        
        Parameters
        ----------
        yaml_dict : dict
            Parsed YAML dictionary from .yaml file.
        """

        yaml_dict = yaml.load(yaml_string, Loader=yaml.BaseLoader)

        # Parse the mandatory parameters
        try:
            self.name = yaml_dict['name']
            self.desc = yaml_dict['desc']

            self.__load_glob(yaml_dict=yaml_dict) # Load globals via a loader function

            if type(self.name) is not str:
                raise TypeError("name is not a string")
            if type(self.desc) is not str:
                raise TypeError("desc is not a string")

        except KeyError as kerr:
            raise fruitError(
                "Rotten fruit! The field {} is not found.".format(str(kerr)))
        except TypeError as te:
            raise fruitError(
                "Rotten fruit! The field {}.".format(str(te)))

        # Global variable substitution
        for each_glob in self.glob:
            yaml_string = yaml_string.replace(
                config.GLOBAL_TOKEN_FMT.format(name=each_glob),
                self.glob[each_glob]
            )
        
        # Reparse yaml for globals in lists
        yaml_dict = yaml.load(yaml_string, Loader=yaml.BaseLoader)

        # Reload the global variables after the substitution
        self.__load_glob(yaml_dict=yaml_dict)

        self.__load_lists(yaml_dict=yaml_dict)

        # List expansions
        for each_list in self.lists:
            yaml_string = yaml_string.replace(
                config.GLOBAL_EXPAND_FMT.format(name=each_list),
                " ".join(self.lists[each_list])
            )

        # Parse yaml again
        yaml_dict = yaml.load(yaml_string, Loader=yaml.BaseLoader)

        self.jobs = []

        # TODO: Verify unique job names
        for each_job in yaml_dict['jobs']:
            jobname = each_job
            jobyaml = yaml_dict['jobs'][jobname]
            self.jobs.append(job.create(jobname, jobyaml, self.execute))

    def __load_glob(self, yaml_dict: dict):
        """
        Load the global variables into `self.glob` when the field is found.
        
        Parameters
        ----------
        `yaml_dict` : dict
            YAML dictionary of the target
        """

        if 'glob' in yaml_dict.keys():
            glob = yaml_dict['glob']

            if type(glob) is not dict:
                raise TargetParseError("The field glob must be a dictionary!")

            for each_key in glob:
                if type(glob[each_key]) is not str:
                    raise TargetParseError("The glob field is not formatted correctly!")
            
            # All test were successful
            self.glob = glob
        else:
            # When the field is nonexistent, use an empty dict
            self.glob = {}

    def __load_lists(self, yaml_dict: dict) -> None:
        """
        Load the lists from the yaml dictionary. Use an empty list when the
        field was left empty.
        
        Parameters
        ----------
        `yaml_dict` : dict
            YAML dictionary of the target
        """
        if 'lists' in yaml_dict.keys():
            lst = yaml_dict['lists']

            if type(lst) is not dict:
                raise TargetParseError("The field lists must be a dictionary!")

            for each_key in lst:
                if type(lst[each_key]) is not list:
                    raise TargetParseError("The lists field is not formatted correctly!")

            self.lists = lst
        else:
            # Initialize with empty list
            self.lists = []
    
    def execute(self, job: str):
        """
        Execute the selected fruit job in fruit.yaml
        
        Parameters
        ----------
        job : str
            Name of the job to execute. job='all' will execute all available jobs
        """

        if job == 'all':
            click.secho('Ripening \'all\' fruits... 🍏🍎')
            click.echo()

            # Execute all jobs
            for each_job in self.jobs:
                each_job.execute()
        else:
            job_flt = filter(lambda j: j.name == job, self.jobs)
            list_jobs = list(job_flt)

            if(len(list_jobs)) < 1:
                raise JobNotFoundError(
                    "The job {job} is not found.".format(job=job))
            else:
                click.secho('Ripening \'{job}\'...'.format(job=job))
                click.echo()

                # Execute the selected job
                list_jobs[0].execute()



"""
Global configuration file for constant global immutable types.
"""

GLOBAL_TOKEN_FMT   = "$({name})"
GLOBAL_EXPAND_FMT  = "$(exp:{name})"
GLOBAL_REPEAT_GMT  = "$(rep:{name})"

FRUIT_FILENAME = "fruit.yaml"

ICON_SUCCESS = "✅" # ICON for showing good execution
ICON_ERROR   = "❌" # ICON for showing execution error
ICON_QUESTION = "❔"

ICON_FRUIT   = "🥝"


LINE_SEP = "-"*80
JOB_SEP  = ">"*80
SHELLCHAR = "$ "


"""
Parent Interface object for pipeline steps
"""
import subprocess
import click
import fruit.config as config
import time
import os

class StepParseError(Exception):
    """Exception for showing errors when parsing steps."""
    pass

class Step(object):
    """
    Interface class for defining a pipeline step.

    Attributes
    ----------
    name: str
        Name of the pipeline step
    type: str
        Type of the pipeline step
    """

    name: str = ""  # Name of the pipeline step
    type: str = ""  # Type of the pipeline step
    __status: int = 1  # Status code of the current step
    _tictime: float = .0  # Start time of the execution
    _toctime: float = .0  # End time of the execution
    _execution_time = .0

    def __init__(self, yaml_dict: dict, exec_callback: callable):
        """
        Create a pipeline step based on a yaml dictionary.
        Parses the `name` and `type` field from the yaml directory and run type checkings.

        Parameters
        ----------
        `yaml_dict` : dict
            YAML dictionary to parse (part of a yaml file)

        `exec_callback`: callable
            Callback from the target to execute a job by name

        Raises
        ------
        NotImplementedError
            The function has to be implemented in subclasses
        """
        try:
            self.name = yaml_dict['name']
            self.type = yaml_dict['type']
        except KeyError as kerr:
            raise StepParseError("The field '{}' cannot be found!".format(str(kerr)))

        if type(self.name) is not str:
            raise StepParseError("The step name has to be a string!")

        if type(self.type) is not str:
            raise StepParseError("The step type has to be a string!")

    @property
    def status(self) -> int:
        """
        Get the execution status of the pipeline step

        Returns
        -------
        int
            Return code of the terminal command
        """
        return self.__status

    @property
    def exec_time(self) -> float:
        """
        Execution time of the current step
        """
        return self._execution_time

    def _tic(self):
        """
        Start a time measurement.
        """
        self._tictime = time.time()
    
    def _toc(self):
        """
        Stop a started time measurement.
        """
        self._toctime = time.time()
        self._execution_time = self._toctime - self._tictime

    def _setstatus(self, status: int):
        """
        Set the execution status of the step after execution
        
        Parameters
        ----------
        status : int
            Returned execution result code from the command
        """
        self.__status = status

    def execute(self) -> bool:
        """
        Execute the current pipeline step and return the result.

        Returns
        -------
        bool
            True = Execution successful; False= Execution error
        
        Raises
        ------
        NotImplementedError
            The function has to be implemented in subclasses
        """
        raise NotImplementedError


class DefaultStep(Step):

    cmd: str = "" # Command string to execute
    args: list = None

    def __init__(self, yaml_dict: dict, exec_callback: callable):
        
        # Initialize the base class for mandatory fields
        super().__init__(yaml_dict=yaml_dict, exec_callback=exec_callback)

        try:
            self.cmd  = yaml_dict['cmd']
            self.args = yaml_dict['args']
        except KeyError as kerr:
            raise StepParseError("The field '{}' cannot be found!".format(str(kerr)))

        if type(self.cmd) is not str:
            raise StepParseError("The cmd field has to be a string!")

        if type(self.args) is not list:
            raise StepParseError("The args field has to be the list!")
    
    def execute(self) -> int:
        """
        Execute the command with the default type handler (`os.system()`)

        Returns
        -------
        int
            Return code of the command execution
        
        Note
        ----
        The given commands are executed via the subprocess modules
        """
        cmdarray = [self.cmd]
        cmdarray.extend(self.args)

        cmdstr = " ".join(cmdarray)

        # Print the command
        click.echo(config.SHELLCHAR + cmdstr)

        try:
            self._tic() # Measure the execution time

            p = subprocess.Popen([cmdstr], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            (child_stdout) = (p.stdout)

            # Set the returncode to 0, when no exception happened
            self._setstatus(0)
        except subprocess.CalledProcessError as cmderror:
            self._setstatus(cmderror.returncode)

        output = child_stdout.read().decode('utf-8')

        # Print the command results
        click.echo(output)

        # NOTE: The time measurement is only accurate after the buffer read!
        self._toc()  # Stop the time measurement

        return self.status

class SystemStep(Step):

    cmd: str = "" # Command string to execute
    args: list = None

    def __init__(self, yaml_dict: dict, exec_callback: callable):

        # Initialize the base class for mandatory fields
        super().__init__(yaml_dict=yaml_dict, exec_callback=exec_callback)

        try:
            self.cmd  = yaml_dict['cmd']
            self.args = yaml_dict['args']
        except KeyError as kerr:
            raise StepParseError("The field '{}' cannot be found!".format(str(kerr)))

        if type(self.cmd) is not str:
            raise StepParseError("The cmd field has to be a string!")

        if type(self.args) is not list:
            raise StepParseError("The args field has to be the list!")
    
    def execute(self) -> int:
        """
        Execute the command with the default type handler (`os.system()`)

        Returns
        -------
        int
            Return code of the command execution
        
        Note
        ----
        The given commands are executed via the subprocess modules
        """
        cmdarray = [self.cmd]
        cmdarray.extend(self.args)

        cmdstr = " ".join(cmdarray)

        # Print the command
        click.echo(config.SHELLCHAR + cmdstr)
        self._tic()
        
        # Run the command in a system subshell
        returncode = os.system(cmdstr)
        self._setstatus(returncode)

        # NOTE: The time measurement is only accurate after the buffer read!
        self._toc()  # Stop the time measurement

        return self.status

class JobStep(Step):
    """
    JobStep object for calling a job from the current yaml file.

    Arguments
    ---------
    job: str
        Name of the job to call
    """

    job: str = "" # Name of the job to call
    __exec_callback: callable = None

    def __init__(self, yaml_dict: dict, exec_callback: callable):

        super().__init__(yaml_dict=yaml_dict, exec_callback=exec_callback)
        try:
            self.job  = yaml_dict['job']

            # Set the execution callback
            self.__exec_callback = exec_callback
        except KeyError as kerr:
            raise StepParseError("The field '{field}' cannot be found".format(field=str(kerr)))

        if type(self.name) is not str:
            raise StepParseError("The step name has to be a string!")
        

    def execute(self):
        """
        Execute the command with the default type handler (`os.system()`)

        Returns
        -------
        int
            Return code of the command execution
        """
        self._tic()
        # Use the callback from the target class to execute the job
        self.__exec_callback(self.job)

        self._toc()

        # TODO: Get execution results
        return 0
        

def create(yaml_dict: dict, exec_callback: callable) -> Step:
    """
    Create a pipeline step object from the yaml dictionary. The creation uses the Factory
    pattern to create an interface typed object from the given representation
    
    Parameters
    ----------
    `yaml_dict` : dict
        YAML dictionary of the pipeline step
    
    `exec_callback`: callabke
        Execution callback from containing target object
    
    Returns
    -------
    Step
        Pipeline step as the Step interface
    """
    if yaml_dict['type'] == 'default':
        return DefaultStep(yaml_dict, exec_callback)
    elif yaml_dict['type'] == 'system':
        return SystemStep(yaml_dict, exec_callback)
    elif yaml_dict['type'] == 'job':
        return JobStep(yaml_dict, exec_callback)
    else:
        raise ValueError("The pipeline type is not available!") 


"""
Implementations of multiple jobs in a yaml file
"""

import fruit.modules.step as step
import click
import fruit.config as config

class JobParseError(Exception):
    """Exception Type for error while parsing fruit jobs"""
    pass

class Job(object):
    """
    Object representation of a job inside of a yaml file

    Attributes:
    -----------

    name: str
        Name of the job
    desc: str
        Description of the job
    steps: list
        Steps of the job
    """

    name: str = ""
    desc: str = "" 
    steps: list = None
    __exec_callback: callable = None

    def __init__(self, name: str, yaml_dict: dict, exec_callback: callable):
        """
        Initialize a job from a parsed yaml dictionary with the default parameters.
        
        Parameters
        ----------
        `name` : str
            Name of the job (parsed in target)
        `yaml_dict` : dict
            Yaml dictionary of the job
        `exec_callback` : callable
            Callback to execute job by bane (passed from `Target`)
        
        Raises
        ------
        JobParseError
            YAML syntax error or missing fields
        """
        try:
            # Fetch mandatory parameters

            self.name = name
            self.desc = yaml_dict['desc']
            self.steps = []

            # Execution callback to target
            self.__exec_callback = exec_callback;

            # Validate Types
            if type(self.name) is not str:
                raise JobParseError("The job name must be a string!")
            
            if type(self.desc) is not str:
                raise JobParseError("The job description must be a string!")

            if type(yaml_dict['steps']) is not list:
                raise JobParseError("The job steps must be formatted as a list!")

            for each_step in yaml_dict['steps']:
                # Fetch the pipeline steps
                self.steps.append(step.create(yaml_dict=each_step, exec_callback=exec_callback))

        except KeyError as kerr:
            raise JobParseError("The field '{}' cannot be found!".format(str(kerr)))
        except JobParseError: # Forward JobParseError errors
            raise
        except step.StepParseError: # Forward step parsing errors
            raise
            

    def execute(self):
        """
        Execute all pipeline steps in the given job. Execution policy may differ by step
        implementation.
        """
        # Print two line separators at each start of a job
        click.echo()
        click.secho(config.JOB_SEP)
        click.secho(config.JOB_SEP)

        number = 0
        for each_step in self.steps:
            # Log execution
            number += 1
            click.secho(config.ICON_FRUIT + " Step {number}: Executing {name} ({type})".format(number=number,name=each_step.name, type=each_step.type), fg="yellow")
            click.secho(config.LINE_SEP)
            
            each_step.execute()
        
        # Print the results after execution
        self.print_results()

        # Print two line separators after each job close
        click.secho(config.JOB_SEP)
        click.secho(config.JOB_SEP)
        click.echo()
    
    def print_results(self):
        """
        Print the results of the pipeline execution steps
        """
        fmt  = "{icon:<3}{time:<7}{name:<}"
        head = fmt.format(icon=config.ICON_QUESTION,time="Time", name="Step")
        sep  = fmt.format(icon="-"*3, time="-"*7,  name="-"*10)

        click.echo() # Make 1 row space
        click.secho(head)
        click.secho(sep)

        for each_step in self.steps:
            if each_step.status == 0:
                click.secho(fmt.format(icon=config.ICON_SUCCESS,time="{:.2f}".format(each_step.exec_time), name=each_step.name))
            else:
                click.secho(fmt.format(icon=config.ICON_ERROR,time="{:.2f}".format(each_step.exec_time), name=each_step.name))




def create(name: str, yaml_dict: dict, exec_callback: callable) -> Job:
    """
    Create a new fruit job
    
    Parameters
    ----------
    name: str
        Name of the job
    yaml_dict : dict
        Dictionary of the job
    
    Returns
    -------
    Job
        Job object
    """

    return Job(name=name, yaml_dict=yaml_dict, exec_callback=exec_callback)


import click
import os
import fruit.config as config
from tabulate import tabulate
from fruit.modules.finder import find_fruit
from fruit.modules.target import Target, JobNotFoundError

@click.group()
def cli():
    """
    List and execute command 'jobs' from fruit.yaml configuration files.
    """
    pass

@cli.command()
@click.argument('path', required=False)
def list(path: str = None):
    r"""
    List all available jobs
    """
    if path is None:
        path = '.'

    try:
        # Find the fruit configuraiton file to execute
        fruit_yaml = find_fruit(path)

        # Parse the fruit target
        target = Target(fruit_yaml)

        try:
            tbllist = []
            # List the available jobs
            for each_job in target.jobs:
                tbllist.append((each_job.name, each_job.desc))
            
            # Print the name and the description of the project
            click.echo()
            click.echo("Target name: {name}".format(name=target.name))
            click.echo("Description: {desc}".format(desc=target.desc))
            click.echo()

            click.echo(tabulate(tbllist, ('Job', 'Description'), tablefmt="simple"))

        except JobNotFoundError as exc:
            click.secho(str(exc), fg='red')

    except FileNotFoundError as exc:
        # The fruit configuration is not found
        click.secho(str(exc), fg='red')

@cli.command()
@click.argument('args', nargs=-1, required=True)
def run(args: list):
    """
    Execute an available job.

    Examples::
    
    \b
        fruit run [JOB]
        fruit run [PATH] [JOB]
    
    PATH - Path (directory) of the configuration file
    JOB  - Name of the job (Use 'all' to run all jobs)
    """

    if len(args) == 1:
        path = '.' # Set the local path
        job  = args[0] # Set the job from argument
    elif len(args) == 2:
        path = args[0] # Set the path from argument
        job = args[1]  # Set the job from the second argument

    # TODO: Add command options for output hiding, etc...

    try:
        # Find the fruit configuraiton file to execute
        fruit_yaml = find_fruit(path)

        # Parse the fruit target
        target = Target(fruit_yaml)

        try:
            target.execute(job=job)
        except JobNotFoundError as exc:
            click.secho(str(exc), fg='red')

    except FileNotFoundError as exc:
        # The fruit configuration is not found
        click.secho(str(exc), fg='red')

@cli.command()
@click.argument('path', required=False)
def edit(path: str):
    r"""Edit the fruit.yaml configuration"""
    if path is None:
        path = "./fruit.yaml"

    if os.path.exists(path) and os.path.isfile(path):
        click.edit(filename=path)
    else:
        click.secho("There is no build configuration to be found. Please create a fruit.yaml file.", fg='red')

if __name__ == '__main__':
    cli()
    


from setuptools import setup, find_packages
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="fruit",
    version="0.1.8",
    packages=find_packages(),
    install_requires=['click', 'colorama', 'pyyaml', 'checksumdir'],
    entry_points={
        'console_scripts': ['fruit=fruit.fruit:cli'],
    },
    author="Marcell Pigniczki",
    author_email="marcip97@gmail.com",
    description="Create and run automations based on yaml files",
    long_description=read('readme.rst'),
    license="MIT",
    keywords="fruit automation yaml",
    url="https://github.com/codesaurus97/fruit",
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3"
        ],
    zip_safe=False,
    package_data={"fruit": ["__init__.py"]},
    python_requires='>=3.6',
)



"""
Module for finding fruit configuration files automatically.
"""

import os

import fruit.config as config


def find_fruit(path: str) -> str:
    """
    Find the fruit.yaml in the given directory and read it's contents.
    
    Parameters
    ----------
    path : str
        Path to search
    
    Returns
    -------
    str
        Contents of the fruit.yaml file
    
    Raises
    ------
    FileNotFoundError
        The given path is not valid
    FileNotFoundError
        The given path does not contain a fruit.yaml file
    """
    if os.path.exists(path) and os.path.isdir(path):
        fruitpath = os.path.join(path, config.FRUIT_FILENAME)

        if os.path.exists(fruitpath) and os.path.isfile(fruitpath):
            
            with open(fruitpath, 'r') as fp:
                fruitfile = fp.read()
                return fruitfile
        else:
            raise FileNotFoundError("There is no build configuration to be found. Please create a fruit.yaml file.")
    else:
        raise FileNotFoundError("The given path is not valid!")
